/*-
 * Copyright (C) 2007 Erik Larsson
 * 
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package org.catacombae.findduplicates;

import java.util.*;
import java.io.*;
import java.security.*;

public class FindDuplicatesEngine {
    public static class Result {
	public final List<Entry> matchingEntries;
	public Result() { matchingEntries = new Vector<Entry>(); }
	public Result(List<Entry> premade) { matchingEntries = premade; }
    }
    public static class Entry {
	public File file;
	public long size;
	public byte[] hash; // Will be null if it hasn't been hashed
	public Entry(File f) { file = f; size = f.length(); hash = null; }
    }
    
    private static boolean abort = false; // Wrapping this inside another object makes it possible to have one abort variable per run of findDuplicates, thus avoiding that any persisting ProgressListener accidentially modifies the state of the abort variable for an unrelated call to findDuplicates. Fix this when possible.
    
    public static Result[] findDuplicates(File[] files, String hashAlgorithm, ProgressListener progressListener) {
	abort = false;
	progressListener.setCancelHandler(new ProgressListener.CancelHandler() {
		public void actionCancel() {
		    abort = true;
		}
	    });

	List<Result> sizeMatches;
	if(true) {
	    ArrayList<File> fileSet = new ArrayList<File>(files.length);
	    for(File f : files)
		fileSet.add(f);
	    Hashtable<Long, LinkedList<File>> bucketTable = new Hashtable<Long, LinkedList<File>>();
	    LinkedList<LinkedList<File>> bucketList = new LinkedList<LinkedList<File>>();
	    
	    for(File f : fileSet) {
		Long currentKey = f.length();
		LinkedList<File> currentBucket = bucketTable.get(currentKey);
		if(currentBucket == null) {
		    currentBucket = new LinkedList<File>();
		    bucketTable.put(currentKey, currentBucket);
		    bucketList.addLast(currentBucket);
		}
		currentBucket.addLast(f);
	    }
	    
	    LinkedList<Result> sizeMatchesTemp = new LinkedList<Result>();
	    for(LinkedList<File> llf : bucketList) {
		if(llf.size() > 1) {
		    Result r = new Result(new LinkedList<Entry>());
		    for(File f : llf)
			r.matchingEntries.add(new Entry(f));
		    sizeMatchesTemp.addLast(r);
		}
	    }
	    sizeMatches = sizeMatchesTemp;
	}
	else {
	    // old and inefficient method
	    Vector<File> fileSet = new Vector<File>(files.length);
	    for(File f : files)
		fileSet.add(f);
	    sizeMatches = new Vector<Result>();
	    /*
	     * For each file in the set:
	     *   Loop through the other files in the set and see if its size matches any of them,
	     *   and if it does, create a new Result and clump them together inside it. Then add
	     *   the result to sizeMatches.
	     */
	    // Note to self: Using a Vector for this purpose introduces a performance bottleneck when removing from set...
	    for(int i = 0; i < fileSet.size(); ++i) {
		if(abort) return null;
		File currentFile = fileSet.get(i);
		Result r = new Result();
		r.matchingEntries.add(new Entry(currentFile));
		if(progressListener != null) progressListener.reportSizeMatchProgress(i/(double)fileSet.size());
		for(int j = i+1; j < fileSet.size(); ++j) {
		    //System.out.println("(" + i + ", " + j + ")");
		    if(abort) return null;
		    File otherFile = fileSet.get(j);
		    if(currentFile.length() == otherFile.length()) {
			r.matchingEntries.add(new Entry(otherFile));
			fileSet.remove(j);
			--j;
		    }
		}
		if(r.matchingEntries.size() > 1)
		    sizeMatches.add(r);
		
	    }
	    if(abort) return null;
	}

	/* Calculate the amount of data that needs to be hashed. */
	long hashedBytes = 0;
	long bytesToHash = 0;
	for(Result r : sizeMatches) {
	    for(Entry e : r.matchingEntries)
		bytesToHash += e.file.length();
	}
	
	/*
	 * We now have to calculate hash values for those entries that are clumped together in
	 * sizeMatches (since if their sizes are identical, we have to find out whether the files
	 * themselves are identical, which is best done by hashing the files and comparing the
	 * hashes).
	 */
	MessageDigest md = null;
	byte[] buffer = new byte[8192];
	int i = 0;
	LinkedList<Result> finalResults = new LinkedList<Result>();
	for(Result r : sizeMatches) {
	    if(abort) return null;
	    if(md == null) {
		try {
		    md = MessageDigest.getInstance(hashAlgorithm);
		} catch(NoSuchAlgorithmException nsae) {
		    throw new RuntimeException("Algorithm " + hashAlgorithm + " not found on system!");
		}
	    }
// 	    double currentHashProgress = i++ / (double)sizeMatches.size();
// 	    double nextHashProgress = (i+1) / (double)sizeMatches.size();
// 	    double currentRange = nextHashProgress - currentHashProgress;
	    if(progressListener != null) progressListener.reportHashProgress(hashedBytes / (double)bytesToHash);
	    
	    // Calculate hashes
	    for(Entry e : r.matchingEntries) {
		if(abort) return null;
		FileInputStream fi = null;
		long bytesRead = 0, fileSize = e.size;
		try {
		    fi = new FileInputStream(e.file);
		    md.reset();
		    while(bytesRead < fileSize) {
			if(abort) {
			    fi.close();
			    return null;
			}
			int curBytesRead = fi.read(buffer);
			if(curBytesRead != -1) {
			    bytesRead += curBytesRead;
			    md.update(buffer, 0, curBytesRead);
			    hashedBytes += curBytesRead;
			}
			else {
			    System.err.println("WARNING: File " + e.file + " seems to have been updated during run time...");
			    break;
			}
// 			double percentageDone = bytesRead / (double)fileSize;
// 			if(progressListener != null) progressListener.reportHashProgress(currentHashProgress+percentageDone*currentRange);
			if(progressListener != null) progressListener.reportHashProgress(hashedBytes / (double)bytesToHash);
			
		    }
		    fi.close(); //glmde!
		    e.hash = md.digest();
		} catch(IOException ioe) {
		    if(fi != null) {
			try { fi.close(); }
			catch(Exception e2) {}
		    }
		    if(progressListener != null) progressListener.reportNonCriticalError(ioe.toString());
		    hashedBytes += fileSize-bytesRead;
		    if(progressListener != null) progressListener.reportHashProgress(hashedBytes / (double)bytesToHash);
		}
	    }
	    if(abort) return null;

	    // Check which ones match
	    for(int j = 0; j < r.matchingEntries.size(); ++j) {
		Entry jentry = r.matchingEntries.get(j);
		Result newr = null;
		for(int k = j+1; k < r.matchingEntries.size(); ++k) {
		    Entry kentry = r.matchingEntries.get(k);
		    if(equals(jentry.hash, kentry.hash)) {
			if(newr == null) {
			    newr = new Result();
			    newr.matchingEntries.add(jentry);
			}
			newr.matchingEntries.add(kentry);
			r.matchingEntries.remove(k);
			--k;
		    }
		}
		if(newr != null)
		    finalResults.add(newr);
		
	    }
	}
	return finalResults.toArray(new Result[finalResults.size()]);
    }
    
    /**
     * Processes all files in selection that are directories, recursively
     * resolving the entire directory tree under the directory.
     */
    public static File[] resolveDirectories(File[] selection, ProgressListener progressListener) {
	abort = false;
	if(progressListener != null)
	    progressListener.setCancelHandler(new ProgressListener.CancelHandler() {
		    public void actionCancel() {
			abort = true;
		    }
		});
	LinkedList<File> result = new LinkedList<File>();
	resolveDirectoriesInternal(result, selection, progressListener, 0.0, 1.0);
	if(abort) return null;
	else return result.toArray(new File[result.size()]);
    }
    private static void resolveDirectoriesInternal(LinkedList<File> result, File[] files, ProgressListener pListener, double progress, double range) {
	double subrange = range/files.length;
	for(File f : files) {
	    if(abort)
		return;// new File[0];
	    try {
		if(f.isDirectory() && f.getCanonicalPath().equals(f.getAbsolutePath())) {
		    //System.err.println("Resolving " + f);
		    File[] subdirContents = f.listFiles();
		    if(subdirContents != null)
			resolveDirectoriesInternal(result, subdirContents, pListener, progress, subrange);
		    else
			System.err.println("Could not read contents of \"" + f + "\"");
		}
		else if(f.isFile())
		    result.add(f);
	    } catch(IOException e) { e.printStackTrace(); }
	    progress += subrange;
	    pListener.reportGatherFilesProgress(progress);
	}
	//return result.toArray(new File[result.size()]);
    }

    private static void resolveTree(File dir, LinkedList<File> result) {
	File[] dirContents = dir.listFiles();
	for(File f : dirContents) {
	    if(abort)
		return;// new File[0];
	    if(f.isDirectory())
		resolveTree(f, result);
	    else
		result.add(f);
	}
    }
    
    private static boolean equals(byte[] a, byte[] b) {
	if(a == null || b == null)
	    return false;
	if(a.length != b.length)
	    return false;
	else {
	    for(int i = 0; i < a.length; ++i)
		if(a[i] != b[i])
		    return false;
	    return true;
	}
    }
}
